from __future__ import unicode_literals


class RulesLightException(Exception):
    """ Base class for all exceptions of this package. """
    pass


class Denied(RulesLightException):
    def __init__(self, rule_text):
        super(Denied, self).__init__(u'%s evaluates to False' % rule_text)


class DoesNotExist(RulesLightException):
    def __init__(self, name):
        super(DoesNotExist, self).__init__(
            u'Rule "%s" is not registered' % name)



from __future__ import unicode_literals
import unittest

import rules_light

@rules_light.make_decorator
def return_true(*args):
    return True


@rules_light.make_decorator
def return_false(*args):
    return False


class DecoratorsTestCase(unittest.TestCase):
    def test_decorator_return_true(self):
        rule = return_true(lambda x, y: True)
        self.assertTrue(rule('x', 'y'))

        rule = return_true(lambda x, y: False)
        self.assertFalse(rule('x', 'y'))

    def test_decorator_return_false(self):
        rule = return_false(lambda x, y: True)
        self.assertFalse(rule('x', 'y'))

        rule = return_false(lambda x, y: False)
        self.assertFalse(rule('x', 'y'))

    def test_decorator_name(self):
        @return_true
        def foo(*args):
            pass

        self.assertEqual(foo.__name__, 'foo')



# -*- encoding: utf-8 -*-
from __future__ import unicode_literals
import logging
import unittest
from mock import Mock

from django.contrib.auth.models import User

import rules_light


class RegistryTestCase(unittest.TestCase):
    def setUp(self):
        self.registry = rules_light.RuleRegistry()
        self.registry.logger = Mock(spec_set=['debug', 'info', 'warn', 'error'])
        self.user, c = User.objects.get_or_create(username='test')

    def test_run_rule_no_args(self):
        mock = Mock(return_value=True, spec_set=['__call__'])
        self.registry['x.y.z'] = mock

        self.registry.logger.debug.assert_called_once_with(
            u'[rules_light] "x.y.z" registered with: Mock')

        result = self.registry.run(self.user, 'x.y.z')

        self.registry.logger.info.assert_called_once_with(
            u'[rules_light] Mock(test, "x.y.z") passed')

        self.assertEqual(result, True)
        mock.assert_called_once_with(self.user, 'x.y.z')

    def test_run_rule_with_args(self):
        mock = Mock(return_value=True, spec_set=['__call__'])
        self.registry['x.y.z'] = mock

        result = self.registry.run(self.user, 'x.y.z', 'foo', x='bar')

        self.registry.logger.info.assert_called_once_with(
            u'[rules_light] Mock(test, "x.y.z", "foo", x="bar") passed')

        self.assertEqual(result, True)
        mock.assert_called_once_with(self.user, 'x.y.z', 'foo', x='bar')

    def test_raises_Denied(self):
        mock = Mock(return_value=False, spec_set=['__call__'])
        self.registry['x.y.z'] = mock

        with self.assertRaises(rules_light.Denied) as cm:
            self.registry.require(self.user, 'x.y.z')

        self.registry.logger.warn.assert_called_once_with(
            u'[rules_light] Deny Mock(test, "x.y.z")')

    def test_return_False(self):
        mock = Mock(return_value=False, spec_set=['__call__'])
        self.registry['x.y.z'] = mock

        self.assertFalse(self.registry.run(self.user, 'x.y.z'))
        self.registry.logger.info.assert_called_once_with(
            u'[rules_light] Mock(test, "x.y.z") failed')

    def test_raises_RuleDoesNotExist(self):
        with self.assertRaises(rules_light.DoesNotExist) as cm:
            self.registry.run(self.user, 'x')

        self.registry.logger.error.assert_called_once_with(
            u'[rules_light] Rule does not exist "x"')



import inspect

from django import template

from classytags.core import Options
from classytags.helpers import AsTag
from classytags.arguments import (Argument, MultiKeywordArgument,
                                  MultiValueArgument)

import rules_light

register = template.Library()


@register.filter
def rule_code(registry, name):
    rule = registry[name]

    if hasattr(rule, '__call__'):
        return inspect.getsource(rule)

    return repr(rule)


class Rule(AsTag):
    options = Options(
        Argument('rule_name'),
        MultiValueArgument('args', required=False),
        MultiKeywordArgument('kwargs', required=False),
        'as',
        Argument('varname', resolve=False, required=False),
    )

    def get_value(self, context, rule_name, args, kwargs):
        return rules_light.run(context['request'].user, rule_name, *args,
                               **kwargs)

register.tag(Rule)



from __future__ import unicode_literals

from django.views import generic
from django.contrib.auth.models import User

import rules_light

@rules_light.class_decorator
class CreateView(generic.CreateView):
    model = User


@rules_light.class_decorator
class UpdateView(generic.UpdateView):
    model = User


@rules_light.class_decorator
class DetailView(generic.DetailView):
    model = User


@rules_light.class_decorator
class DeleteView(generic.DeleteView):
    model = User


@rules_light.class_decorator('funny')
class FunnyUpdateView(generic.UpdateView):
    model = User






"""
It is trivial to take shortcuts because the rule registry is a simple dict.

You can reuse your rules several times in standard python::

    def my_model_or_is_staff(user, rule, model, obj=None):
        return user.is_staff or (obj and obj.author == user)

    rules_light.registry.setdefault('your_app.your_model.create',
        my_model_or_is_staff)
    rules_light.registry.setdefault('your_app.your_model.update',
        my_model_or_is_staff)
    rules_light.registry.setdefault('your_app.your_model.delete',
        my_model_or_is_staff)

This module provides some shortcut(s). Shortcuts are also usable as decorators
too (see ``make_decorator``)::

    @rules_light.is_authenticated
    def my_book(user, rule, book):
        return book.author == user

    rules_light.registry.setdefault('your_app.your_model.update', my_book)
"""
from __future__ import unicode_literals

from .decorators import make_decorator

__all__ = ['is_staff', 'is_authenticated']


@make_decorator
def is_staff(user, rulename, *args, **kwargs):
    """
    Return True if user.is_staff.

    For example, in ``your_app/rules_light_registry.py``::

        rules_light.registry.setdefault('your_app.your_model.create',
            rules_light.is_staff)

    Is equivalent to::

        rules_light.registry.setdefault('your_app.your_model.create',
            lambda: user, rulename, *args, **kwargs: user.is_staff)

    Also::

        rules_light.registry.setdefault('your_app.your_model.create',
            rules_light.is_staff(your_rule_callback))

    Is equivalent to::

        def staff_and_stuff(user, rule, *args, **kwargs):
            if not rules_light.is_staff(user, rule, *args, **kwargs):
                return False

            if your_stuff():
                return True

        rules_light.registry.setdefault('your_app.your_model.create',
            rules_light.is_staff(your_rule_callback))
    """
    return user and user.is_staff


@make_decorator
def is_authenticated(user, rulename, *args, **kwargs):
    """
    Return user.is_authenticated().
    """
    return user and user.is_authenticated()



"""
This module enables piling rules on each others.

Consider this simple rule::

    def is_authenticated(user, *args, **kwargs):
        return user and user.is_authenticated()

It can of course be used directly::

    rules_light.registry['do_something'] = is_authenticated

But if defined using ``make_decorator`` as such::

    @rules_light.make_decorator
    def is_authenticated(user, *args, **kwargs):
        return user and user.is_authenticated()

Then you can use it to decorate other rules too::

    @is_authenticated
    def my_book(user, rule, book):
        return user == book.author

    rules_light.registry['do_something'] = my_book

"""
from __future__ import unicode_literals


def make_decorator(_rule):
    def _decorator(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            func = args[0]

            def _decorated(user, rule, *args, **kwargs):
                if not _rule(user, rule, *args, **kwargs):
                    return False
                return func(user, rule, *args, **kwargs)
            _decorated.__name__ = func.__name__
            return _decorated
        else:  # rule
            return _rule(*args, **kwargs)
    _decorator.__name__ = _rule.__name__
    return _decorator



"""
The rule registry is in charge of keeping and executing security rules.

It is the core of this app, everything else is optionnal.

This module provides a variable, ``registry``, which is just a module-level,
default RuleRegistry instance.

A rule can be a callback or a variable that will be evaluated as bool.
"""
from __future__ import unicode_literals

import logging

from django.utils.encoding import smart_text

from .exceptions import Denied, DoesNotExist

__all__ = ('RuleRegistry', 'registry', 'require', 'run', 'autodiscover')


class RuleRegistry(dict):
    """
    Dict subclass to manage rules.

    logger
        The standard logging logger instance to use.
    """
    def __init__(self):
        self.logger = logging.getLogger('rules_light')

    def __setitem__(self, key, value):
        """
        Adds a debug-level log on registration.
        """
        super(RuleRegistry, self).__setitem__(key, value)
        self.logger.debug(u'[rules_light] "%s" registered with: %s' % (
            key, self.rule_text_name(value)))

    def run(self, user, name, *args, **kwargs):
        """
        Run a rule, return True if whatever it returns evaluates to True.

        Also logs calls with the info-level.
        """
        if name not in self:
            self.logger.error(u'[rules_light] Rule does not exist "%s"' % name)
            raise DoesNotExist(name)

        rule = self[name]

        if hasattr(rule, '__call__'):
            result = self[name](user, name, *args, **kwargs)
        else:
            result = rule

        text = self.as_text(user, name, *args, **kwargs)
        if result:
            self.logger.info(u'[rules_light] %s passed' % text)
            return True
        else:
            self.logger.info(u'[rules_light] %s failed' % text)
            return False

    def require(self, user, name, *args, **kwargs):
        """
        Run a rule, raise ``rules_light.Denied`` if returned False.

        Log denials with warn-level.
        """
        result = self.run(user, name, *args, **kwargs)

        if not result:
            text = self.as_text(user, name, *args, **kwargs)
            self.logger.warn(u'[rules_light] Deny %s' % text)
            raise Denied(text)

    def as_text(self, user, name, *args, **kwargs):
        """ Format a rule to be human readable for logging """
        if name not in self:
            raise DoesNotExist(name)

        formated_args = []
        for arg in args:
            formated_args.append(u'"%s"' % smart_text(arg))

        for key, value in kwargs.items():
            formated_args.append(u'%s="%s"' % (smart_text(key),
                smart_text(value)))
        formated_args = u', '.join(formated_args)

        if hasattr(self[name], '__call__'):
            text_name = self.rule_text_name(self[name])

            if formated_args:
                return u'%s(%s, "%s", %s)' % (text_name, user, name,
                                              formated_args)
            else:
                return u'%s(%s, "%s")' % (text_name, user, name)
        else:
            return u'%s is %s' % (name, self[name])

    def rule_text_name(self, rule):
        if hasattr(rule, 'func_name'):
            return rule.func_name
        elif rule is True:
            return u'True'
        elif rule is False:
            return u'False'
        elif hasattr(rule, '__name__'):
            return rule.__name__
        elif hasattr(rule, '__class__'):
            return rule.__class__.__name__
        else:
            return smart_text(rule)


registry = RuleRegistry()


def run(user, name, *args, **kwargs):
    """ Proxy ``rules_light.registry.run()``. """
    return registry.run(user, name, *args, **kwargs)


def require(user, name, *args, **kwargs):
    """ Proxy ``rules_light.registry.require()``. """
    registry.require(user, name, *args, **kwargs)


def _autodiscover(registry):
    """See documentation for autodiscover (without the underscore)"""
    import copy
    from django.conf import settings
    from django.utils.importlib import import_module
    from django.utils.module_loading import module_has_submodule

    for app in settings.INSTALLED_APPS:
        mod = import_module(app)
        # Attempt to import the app's admin module.
        try:
            before_import_registry = copy.copy(registry)
            import_module('%s.rules_light_registry' % app)
        except:
            # Reset the model registry to the state before the last import as
            # this import will have to reoccur on the next request and this
            # could raise NotRegistered and AlreadyRegistered exceptions
            # (see #8245).
            registry = before_import_registry

            # Decide whether to bubble up this error. If the app just
            # doesn't have an admin module, we can ignore the error
            # attempting to import it, otherwise we want it to bubble up.
            if module_has_submodule(mod, 'rules_light_registry'):
                raise


def autodiscover():
    """
    Check all apps in INSTALLED_APPS for stuff related to rules_light.

    For each app, autodiscover imports ``app.rules_light_registry`` if
    available, resulting in execution of ``rules_light.registry[...] = ...``
    statements in that module, filling registry.

    Consider a standard app called 'cities_light' with such a structure::

        cities_light/
            __init__.py
            models.py
            urls.py
            views.py
            rules_light_registry.py

    With such a rules_light_registry.py::

        import rules_light

        rules_light.register('cities_light.city.read', True)
        rules_light.register('cities_light.city.update',
            lambda user, rulename, country: user.is_staff)

    When autodiscover() imports cities_light.rules_light_registry, both
    `'cities_light.city.read'` and `'cities_light.city.update'` will be
    registered.
    """
    _autodiscover(registry)



"""
The role of the middleware is to present a user friendly error page when a rule
denied process of the request by raising ``Denied``.
"""
from __future__ import unicode_literals

from django import template
from django import http
from django.conf import settings

from .exceptions import RulesLightException, Denied, DoesNotExist


class Middleware(object):
    """
    Install this middleware by adding `rules_light.middleware.Middleware`` to
    ``settings.MIDDLEWARE_CLASSES``.
    """
    def process_exception(self, request, exception):
        """
        Render ``rules_light/exception.html`` when a ``Denied`` exception was
        raised.
        """
        if not isinstance(exception, RulesLightException):
            return

        ctx = template.RequestContext(request, dict(exception=exception,
            settings=settings))
        return http.HttpResponseForbidden(template.loader.render_to_string(
            'rules_light/exception.html', ctx))



"""

"""
from __future__ import unicode_literals
import six

from django.views import generic

from .exceptions import RulesLightException
from .registry import registry


__all__ = ('class_decorator',)


def patch_get_object(cls, suffix, override):
    old_get_object = cls.get_object

    def new_get_object(self, *args, **kwargs):
        obj = old_get_object(self, *args, **kwargs)

        if self.get_object._rule_override:
            rule_name = self.get_object._rule_override
        else:
            rule_name = '%s.%s.%s' % (obj.__class__._meta.app_label,
                obj.__class__._meta.module_name, self.get_object._rule_suffix)

        registry.require(self.request.user, rule_name, obj)

        return obj

    new_get_object._rule_suffix = suffix
    new_get_object._rule_override = override
    cls.get_object = new_get_object


class class_decorator(object):
    """
    Can be used to secure class based views.

    If the view has ``model=YourModel``, it will support:

    - ``CreateView``, it will decorate ``get_form()``, to run
      ``rules_light.require('yourapp.yourmodel.create')``,
    - ``UpdateView``, it will decorate ``get_object()``, to run
      ``rules_light.require('yourapp.yourmodel.update', obj)``,
    - ``DeleteView``, it will decorate ``get_object()``, to run
      ``rules_light.require('yourapp.yourmodel.delete', obj)``,
    - ``DetailView``, it will decorate ``get_object()``, to run
      ``rules_light.require('yourapp.yourmodel.read', obj)``,
    - others views, if the rule name is specified in the decorator for example
      ``@class_decorator('some_rule')``, then it will decorate ``dispatch()``,
    - Else it raises an exception.
    """
    rule = None

    def __new__(self, *args):
        if hasattr(args[0], 'as_view'):
            cls = args[0]
        elif isinstance(args[0], six.string_types):
            if six.PY2:
                decorator_name = b'new_class_decorator'
            elif six.PY3:
                decorator_name = 'new_class_decorator'

            new_class_decorator = type(decorator_name,
                (class_decorator,), {'rule': args[0]})
            return new_class_decorator
        elif hasattr(args[0], '__call__'):
            raise Exception("No function support")
        else:
            raise Exception("What?")

        if issubclass(cls, generic.CreateView):
            old_get_form = cls.get_form

            def new_get_form(self, form_class, *args, **kwargs):
                model = form_class.Meta.model
                rule_name = '%s.%s.create' % (model._meta.app_label,
                    model._meta.module_name)

                registry.require(self.request.user, rule_name)

                return old_get_form(self, form_class, *args, **kwargs)

            cls.get_form = new_get_form

        elif issubclass(cls, generic.UpdateView):
            patch_get_object(cls, 'update', self.rule)

        elif issubclass(cls, generic.DetailView):
            patch_get_object(cls, 'read', self.rule)

        elif issubclass(cls, generic.DeleteView):
            patch_get_object(cls, 'delete', self.rule)

        elif self.rule:
            old_dispatch = cls.dispatch

            def new_dispatch(self, request, *args, **kwargs):
                registry.require(request.user, self.dispatch._rule)
                return old_dispatch(self, request, *args, **kwargs)
            new_dispatch._rule = self.rule
            cls.dispatch = new_dispatch

        else:
            raise RulesLightException('Dont understand what to do')

        return cls



from __future__ import unicode_literals

from django.views import generic

import rules_light


@rules_light.class_decorator('rules_light.rule.read')
class RegistryView(generic.TemplateView):
    """
    Expose the rule registry for debug purposes.

    Install it as such::

        url(r'^rules/$', RegistryView.as_view(), name='rules_light_registry'),

    Or just::

        url(r'^rules/', include('rules_light.urls')),

    Note: view requires ``'rules_light.rule.read'`` which is enabled for admins
    by default.
    """

    template_name = 'rules_light/registry.html'

    def get_context_data(self):
        """ Add the registry to the context. """
        return {'registry': rules_light.registry}



from __future__ import unicode_literals
import unittest

from django.test.client import Client

import rules_light


class MiddlewareTestCase(unittest.TestCase):
    def setUp(self):
        self.client = Client()

    def test_redirect(self):
        self.client.go('/')



from __future__ import unicode_literals

import unittest

import rules_light


class AutodiscoverTestCase(unittest.TestCase):
    def test_autodiscover(self):
        self.assertEqual(list(rules_light.registry.keys()), [])
        rules_light.autodiscover()
        self.assertTrue('rules_light.rule.read' in rules_light.registry.keys())
        self.assertFalse('Foo bar' in rules_light.registry.keys())



from __future__ import unicode_literals
import unittest

from django.views import generic
from django.test.client import RequestFactory
from django.contrib.auth.models import User

import rules_light
from ..views import RegistryView

from .class_decorator_classes import *


class ClassDecoratorTestCase(unittest.TestCase):
    def setUp(self):
        self.request = RequestFactory().get('/')
        self.request.user, c = User.objects.get_or_create(username='foo')

    def test_create_view_decorator(self):
        rules_light.registry['auth.user.create'] = False
        view = CreateView.as_view()

        with self.assertRaises(rules_light.Denied) as cm:
            view(self.request)

        rules_light.registry['auth.user.create'] = True
        # it should not raise an exception
        view(self.request)

    def test_update_view_decorator(self):
        rules_light.registry['auth.user.update'] = False
        view = UpdateView.as_view()

        with self.assertRaises(rules_light.Denied) as cm:
            view(self.request, pk=1)

        rules_light.registry['auth.user.update'] = True
        # it should not raise an exception
        view(self.request, pk=1)

    def test_detail_view_decorator(self):
        rules_light.registry['auth.user.read'] = False
        view = DetailView.as_view()

        with self.assertRaises(rules_light.Denied) as cm:
            view(self.request, pk=1)

        rules_light.registry['auth.user.read'] = True
        # it should not raise an exception
        view(self.request, pk=1)

    def test_delete_view_decorator(self):
        rules_light.registry['auth.user.delete'] = False
        view = DeleteView.as_view()

        with self.assertRaises(rules_light.Denied) as cm:
            view(self.request, pk=1)

        rules_light.registry['auth.user.delete'] = True
        # it should not raise an exception
        view(self.request, pk=1)

    def test_funny_view_decorator(self):
        rules_light.registry['funny'] = False
        # ensure that it would not raise an exception if it tried
        # auth.user.read
        rules_light.registry['auth.user.read'] = True
        view = FunnyUpdateView.as_view()

        with self.assertRaises(rules_light.Denied) as cm:
            view(self.request, pk=1)

        rules_light.registry['funny'] = True
        # it should not raise an exception
        view(self.request, pk=1)

    def test_dispatch_decorator(self):
        rules_light.registry['foo'] = False

        @rules_light.class_decorator('foo')
        class MyView(generic.View):
            pass
        view = MyView.as_view()

        with self.assertRaises(rules_light.Denied) as cm:
            view(self.request)

        rules_light.registry['foo'] = True
        # it should not raise an exception
        view(self.request)

    def test_fail(self):
        with self.assertRaises(rules_light.RulesLightException) as cm:
            @rules_light.class_decorator
            class MyView(generic.View):
                pass



from __future__ import unicode_literals

from django.conf.urls import patterns, include, url

import rules_light

from .views import RegistryView


urlpatterns = patterns('',
    url(r'$', RegistryView.as_view(), name='rules_light_registry'),
)



from __future__ import unicode_literals
import unittest

from django.contrib.auth.models import User

import rules_light


class ShortcutsTestCase(unittest.TestCase):
    def setUp(self):
        self.user, c = User.objects.get_or_create(username='foo')
        self.admin, c = User.objects.get_or_create(username='bar',
                is_staff=True)

    def test_is_authenticated_decorator(self):
        return_true = rules_light.is_authenticated(lambda u, r: True)

        self.assertFalse(return_true(None, 'foo'))
        self.assertTrue(return_true(self.user, 'foo'))

    def test_is_authenticated_rule(self):
        self.assertFalse(rules_light.is_authenticated(None, 'foo'))
        self.assertTrue(rules_light.is_authenticated(self.user, 'foo'))

    def test_is_staff_decorator(self):
        return_true = rules_light.is_staff(lambda u, r: True)

        self.assertFalse(return_true(None, 'foo'))
        self.assertFalse(return_true(self.user, 'foo'))
        self.assertTrue(return_true(self.admin, 'foo'))

    def test_is_staff_rule(self):
        self.assertFalse(rules_light.is_staff(None, 'foo'))
        self.assertFalse(rules_light.is_staff(self.user, 'foo'))
        self.assertTrue(rules_light.is_staff(self.admin, 'foo'))



from __future__ import unicode_literals
import unittest

from django.test.client import RequestFactory
from django.contrib.auth.models import User, AnonymousUser

import rules_light
from ..views import RegistryView


class ViewsTestCase(unittest.TestCase):
    def setUp(self):
        """
        Note that URL doesn't matter here because the tests excute the views
        directly.
        """
        User.objects.all().delete()

        self.anonymous_request = RequestFactory().get('/')
        self.anonymous_request.user = AnonymousUser()

        self.user_request = RequestFactory().get('/')
        self.user_request.user, c = User.objects.get_or_create(
            username='foo', is_staff=False)

        self.admin_request = RequestFactory().get('/')
        self.admin_request.user, c = User.objects.get_or_create(
            username='bar', is_staff=True)

    def test_registry_view(self):
        view = RegistryView.as_view()

        with self.assertRaises(rules_light.Denied) as cm:
            view(self.anonymous_request)

        with self.assertRaises(rules_light.Denied) as cm:
            view(self.user_request)

        # it should not raise an exception
        view(self.admin_request)



from .registry import RegistryTestCase
from .class_decorator import ClassDecoratorTestCase
from .autodiscover import AutodiscoverTestCase
from .views import ViewsTestCase
from .decorators import DecoratorsTestCase
from .shortcuts import ShortcutsTestCase



from .registry import RuleRegistry, registry, run, require, autodiscover
from .class_decorator import class_decorator
from .decorators import make_decorator
from .exceptions import Denied, DoesNotExist, RulesLightException
from .middleware import Middleware
from .shortcuts import is_authenticated, is_staff
